"""AD - Groups

This module contains actions for viewing and managing Azure AD Groups.
"""
import uuid
import logging

from typing import List, Union, Optional, Dict, Any


from azure.graphrbac.models import GroupCreateParameters, CheckGroupMembershipParameters
from azure.graphrbac.models.graph_error_py3 import GraphErrorException

from ppa import ui

from ..exceptions import ObjectNotFound
from ..records import AzureClient
from .._common import get_table_field
from ._ad_api import client, api_call, process_groups, process_objects, process_bool
from .records import Group, User


logger = logging.getLogger(__name__)


GROUP_TABLE_HEADER = ["Display Name", "Description", "Mail Enabled", "Security Enabled"]
GROUP_TABLE_FIELDS = ["display_name", "description", "mail_enabled", "security_enabled"]


@api_call
def get_all(azure_client: AzureClient) -> List[Group]:
    """Get all groups.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

    ??? success "Output"

        A list of [Group](../inputs_outputs/#group) dictionaries.

    ??? example
        ```YAML linenums="1"
        - azure.ad.groups.get_all:
          load:
            azure_client: azure_secrets
          save: all_groups
        ```
    """
    return process_groups(client(azure_client).groups.list())


@api_call
def get_by_name(azure_client: AzureClient, name: str) -> Group:
    """Get the group with the supplied display name.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __name__: the group display name

    ??? success "Output"

        A single [Group](../inputs_outputs/#group) dictionary.

    ??? example
        ```YAML linenums="1"
        - azure.ad.groups.get_by_name:
            name: ppa
          load:
            azure_client: azure_secrets
          save: ppa_group
        ```
    """
    try:
        return process_groups(client(azure_client).groups.list(filter="displayName eq '{}'".format(name)))[0]
    except IndexError:
        message = f"No group was found with name '{name}'."
        ui.output_error(message)
        raise ObjectNotFound(message)


@api_call
def create_security_group(
    azure_client: AzureClient, display_name: str, mail_nickname: str, additional_properties: Optional[Dict[str, Any]] = None,
) -> Group:
    """Create a new security group.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __display_name:__ a display name for the group

        - __mail_nickname:__ a mail alias for the group

        - __additional_properties:__ a dictionary containing any other attributes to set (see [here](https://docs.microsoft.com/en-us/python/api/azure-graphrbac/azure.graphrbac.models.groupcreateparameters?view=azure-python) for more information)

    ??? success "Output"

         A single [Group](../inputs_outputs/#group) dictionary.

    ??? example
        ```YAML linenums="1"
        - azure.ad.groups.create_security_group:
            display_name: Help Desk
            mail_nickname: help.desk
          load:
            azure_client: azure_secrets
          save: new_group
        ```
    """
    # https://docs.microsoft.com/en-us/python/api/azure-graphrbac/azure.graphrbac.models.groupcreateparameters?view=azure-python
    return process_groups(
        client(azure_client).groups.create(
            GroupCreateParameters(
                display_name=display_name,
                mail_nickname=mail_nickname,
                security_enabled=True,
                mail_enabled=False,
                additional_properties=additional_properties if additional_properties else {},
            )
        )
    )[0]


@api_call
def delete_by_id(azure_client: AzureClient, object_id: str) -> None:
    """Delete the group with the supplied object ID.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __object_id:__ the group's object ID

    !!! warning "Delete Permissions"

        Deleting an object requires your API application in Azure to be assigned the Global Administrators role.

    ??? success "Output"

         Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - azure.ad.groups.delete_by_id:
          load:
            object_id: group.object_id
            azure_client: azure_secrets
        ```
    """
    client(azure_client).groups.delete(object_id)


# Not decorated as it calls other decorated functions
def delete_by_name(azure_client: AzureClient, name: str) -> None:
    """Delete the group with the supplied display name.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __name:__ the name of the group to delete

    !!! warning "Delete Permissions"

        Deleting an object requires your API application in Azure to be assigned the Global Administrators role.

    ??? success "Output"

         Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - azure.ad.groups.delete_by_name:
          load:
            name: group.display_name
            azure_client: azure_secrets
        ```
    """
    client(azure_client).groups.delete(get_by_name(azure_client, name)["object_id"])


@api_call
def has_member(azure_client: AzureClient, group_object_id: str, member_object_id: str) -> bool:
    """Determine whether the supplied member is in the supplied group.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __group_object_id:__ the group object ID

        - __member_object_id:__ the member object ID to check

    ??? success "Output"

        A boolean is outputted by this action:

        - `true` if the supplied member is in the group

        - `false` if the supplied member is not in the group

    ??? example
        ```YAML linenums="1"
        - azure.ad.groups.has_member:
          load:
            group_object_id: group.object_id
            member_object_id: member.object_id
          save: is_member
        ```
    """
    return process_bool(
        client(azure_client).groups.is_member_of(
            CheckGroupMembershipParameters(group_id=group_object_id, member_id=member_object_id)
        )
    )


@api_call
def get_members(azure_client: AzureClient, object_id: str) -> List[Union[User, Group]]:
    """Get all members of the supplied group.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __object_id:__ the group object ID

    ??? success "Output"

        A list of [User](../inputs_outputs/#user) or [Group](../inputs_outputs/#group) dictionaries.

    ??? example
        ```YAML linenums="1"
        - azure.ad.groups.get_members:
          load:
            object_id: group.object_id
            azure_client: azure_secrets
          save: member_list
        ```
    """
    return process_objects(client(azure_client).groups.get_group_members(object_id))


@api_call
def get_users(azure_client: AzureClient, object_id: str) -> List[User]:
    """Get all users who are a member of the supplied group.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __object_id:__ the group object ID

    ??? success "Output"

        A list of [User](../inputs_outputs/#user) dictionaries.

    ??? example
        ```YAML linenums="1"
        - azure.ad.groups.get_users:
          load:
            object_id: group.object_id
            azure_client: azure_secrets
          save: user_list
        ```
    """
    return process_objects(
        list(filter(lambda m: m.object_type == "User", client(azure_client).groups.get_group_members(object_id)))
    )


@api_call
def get_groups(azure_client: AzureClient, object_id: str) -> List[Group]:
    """Get all groups that are a member of the supplied group.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __object_id:__ the group object ID

    ??? success "Output"

        A list of [Group](../inputs_outputs/#group) dictionaries.

    ??? example
        ```YAML linenums="1"
        - azure.ad.groups.get_groups:
          load:
            object_id: group.object_id
            azure_client: azure_secrets
          save: group_list
        ```
    """
    return process_objects(
        list(filter(lambda m: m.object_type == "Group", client(azure_client).groups.get_group_members(object_id)))
    )


@api_call
def add_member(azure_client: AzureClient, group_object_id: str, member_object_id: str) -> None:
    """Add a member to a group.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __group_object_id:__ the object ID of the group

        - __member_object_id:__ the object ID of the new member

    ??? success "Output"

        Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - azure.ad.groups.add_member:
          load:
            group_object_id: group.object_id
            member_object_id: user.object_id
            azure_client: azure_secrets
        ```
    """
    try:
        return client(azure_client).groups.add_member(
            group_object_id, f"https://graph.windows.net/{azure_client['tenant_id']}/directoryObjects/{member_object_id}"
        )
    except GraphErrorException as e:
        logger.exception(e)
        error = str(e).lower()
        if "invalid object identifier" in error:
            message = "Failed to add group member as one or more of the supplied object IDs is invalid."
            ui.output_error(message)
            raise ObjectNotFound(message)
        raise e


@api_call
def remove_member(azure_client: AzureClient, group_object_id: str, member_object_id: str) -> None:
    """Remove a member from a group.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __group_object_id:__ the object ID of the group

        - __member_object_id:__ the object ID of the member to remove

    ??? success "Output"

        Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - azure.ad.groups.remove_member:
          load:
            group_object_id: group.object_id
            member_object_id: user.object_id
            azure_client: azure_secrets
        ```
    """
    try:
        return client(azure_client).groups.remove_member(group_object_id, member_object_id)
    except GraphErrorException as e:
        logger.exception(e)
        error = str(e).lower()
        if "invalid object identifier" in error:
            message = "Failed to remove group member as one or more of the supplied object IDs is invalid."
            ui.output_error(message)
            raise ObjectNotFound(message)
        raise e


def output_table(text: str, groups: Union[Group, List[Group]]) -> None:
    """Display a list of groups in a table. The table will have the following columns:

    The table will have the following columns:

    - Display Name
    - Description
    - Mail Enabled
    - Security Enabled

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __text:__ the title of the table

        - __group_list:__ a list of [Group](../inputs_outputs/#group) dictionaries to display in the table

    ??? success "Output"
        Nothing is outputted by this action.

    ??? example
        - Show a list of groups in a table

        ```YAML linenums="1"
        - azure.ad.groups.output_table:
            text: All groups for user: Jon Smith
          load:
            group_list: user_groups
        ```
    """
    groups = [groups] if not isinstance(groups, list) else groups
    ui.output_table(
        text,
        header=GROUP_TABLE_HEADER,
        data=[[get_table_field(group, field) for field in GROUP_TABLE_FIELDS] for group in groups],
    )


def input_table(
    text: str, groups: Union[Group, List[Group]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[Group]:
    """Display groups in a table, & allow the task operator to make a selection.

    The table will have the following columns:

    - Display Name
    - Description
    - Mail Enabled
    - Security Enabled

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __text:__ the title of the table

        - __groups:__ a single or list of [Group](../inputs_outputs/#group) dictionaries to display in the table

        - __minimum:__ the minimum number of acceptable selections

        - __maximum:__ the maximum number of acceptable selections

    ??? success "Output"

        A list of [Group](../inputs_outputs/#group) dictionaries.

    ??? tip

        - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.
        - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number
            of selections.

    ??? example

        ```YAML linenums="1"
        - active_directory.groups.get_all:
          load:
            azure_client: azure_client
          save: all_groups

        - active_directory.groups.input_table:
            text: Please select up to 3 groups
            minimum: 1
            maximum: 3
          load:
            groups: all_groups
          save: selected_groups
        ```
    """
    groups = [groups] if not isinstance(groups, list) else groups
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text,
            header=GROUP_TABLE_HEADER,
            data=[[get_table_field(group, field) for field in GROUP_TABLE_FIELDS] for group in groups],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [groups[i] for i in indices]
