"""Management - Resource Groups

This module contains actions for viewing Azure resource groups.
"""

from typing import List, Optional
from ._resource_api import client, process_resource_groups
from .records import ResourceGroup
from ..records import AzureClient
from .._common import input_table as common_input_table


def get_all(azure_client: AzureClient, subscription_id: str) -> List[ResourceGroup]:
    """Get all resource groups.

    !!! question "Minimum Plugin Version: 1.1.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __subscription_id__: the subscription ID

    ??? success "Output"
        A list of [ResourceGroup](../inputs_outputs/#resourcegroup) dictionaries.

    ??? example
        ```YAML linenums="1"
        - azure.mgmt.groups.get_all:
          load:
            azure_client: azure_secrets
            subscription_id: subscription_id
          save: all_groups
        ```
    """

    return process_resource_groups(client(client=azure_client, subscription_id=subscription_id).resource_groups.list())


def input_table(
    text: str, groups: List[ResourceGroup], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[ResourceGroup]:
    """Display a list of resource groups in a table, & allow the task operator to make a selection.

    The table will have the following columns:

    - Name
    - Location
    - Tags

    !!! question "Minimum Plugin Version: 1.1.0"

    ??? info "Input"

        - __text:__ the title of the table

        - __groups:__ a list of [ResourceGroup](../inputs_outputs/#resourcegroup) dictionaries to display in the table

        - __minimum:__ The minimum number of acceptable selections (Optional)

        - __maximum:__ The maximum number of acceptable selections (Optional)

    ??? success "Output"

        A list of [ResourceGroup](../inputs_outputs/#resourcegroup) dictionaries.

    ??? tip

        - If __neither__ a `minimum` _or_ `maximum` is provided, the task operator will be able to submit 0 selections.

        - If `minimum` _or_ `maximum` are provided, the operation will repeat until a valid number of selections is made.

    ??? example

        Finding groups with [get_all](#get_all), saving them as `all_groups`, & waiting for a single selection:

        ```YAML linenums="1" hl_lines="5 12 13"
        - azure.mgmt.groups.get_all:
          load:
            subscription_id: subscription_id
            azure_client: azure_secrets
          save: all_groups

        - azure.mgmt.groups.input_table:
            text: Choose a resource group
            minimum: 1
            maximum: 1
          load:
            groups: all_groups
          save: selected_groups
        ```
    """

    return common_input_table(
        text=text,
        data=groups,
        minimum=minimum,
        maximum=maximum,
        header=["Name", "Location", "Tags"],
        fields=[lambda group: group.get("name"), lambda group: group.get("location"), lambda group: group.get("tags") or ""],
    )
